#
# graphics.py
# UVa cs1120 
# Problem Set 3
#

import turtle
import math

#The following procedures create L-system commands
def make_forward_command():
    return ["F"]
def make_rotate_command(x):
    return ["R",x]
def make_offshoot_command(x):
    return ["O",x]

def make_lsystem_command(*args):
    return list(args)

# L-System commands for the tree fractal
# This is the replacement
# F1 ::= (F1 O(R30 F1) F1 O(R-60 F1) F1)
tree_commands = make_lsystem_command(
    make_forward_command(),
    make_offshoot_command(make_lsystem_command(
        make_rotate_command(30),make_forward_command())),
    make_forward_command(),
    make_offshoot_command(make_lsystem_command(
        make_rotate_command(-60), make_forward_command())),
    make_forward_command())

# Graphics Support
# The word "class" means that we're using Object-Oriented Programming
# here, a concept that we will cover later in the course. Don't
# worry about "class" or "__init__" for now. 
class Graph:
    def __init__(self, width, height):
        self.w = width
        self.h = height
        turtle.setup(self.w,self.h,0,0)
        self.cv = turtle.getcanvas()

g = Graph(600,600) # Default window size is 600x600, can be changed
s = turtle.Screen()

def kill():
    s.bye()
s.onkey(kill,"q")

# Making points:
def make_point(x,y):
    return [x,y,(0,0,0)]

def make_colored_point(x,y,c):
    return [x,y,c]

def x_of_point(point):
    return point[0]
def y_of_point(point):
    return point[1]
def color_of_point(point):
    return point[2]

# Support for working with colors:
def dec_to_hex(val):
    v = hex(val)[2:]
    if len(v) < 2:
        v = "0" + v
    return v

def convert_color_to_string(c):
    return "#" + dec_to_hex(c[0]) + dec_to_hex(c[1]) + dec_to_hex(c[2])

# Drawing points and lines
def window_draw_point(point):
    new_x = int(point[0]*g.w-g.h/2)
    new_y = -1*int(point[1]*g.h-g.h/2)
    cl = convert_color_to_string(point[2])
    g.cv.create_line(new_x,new_y,new_x+1,new_y-1,fill=cl)
    turtle.update()

def window_draw_line(point1,point2):
    new_x_1 = int(point1[0]*g.w-g.h/2)
    new_y_1 = -1*int(point1[1]*g.h-g.h/2)
    new_x_2 = int(point2[0]*g.w-g.h/2)
    new_y_2 = -1*int(point2[1]*g.h-g.h/2)
    g.cv.create_line(new_x_1,new_y_1,new_x_2,new_y_2)
    turtle.update()

# Drawing a curve

def draw_curve_points(curve,n):
        for step in range(n):
                time = 1.0 * step / n
                window_draw_point(curve(time))


def draw_curve_connected(curve,n):
        for step in range(n-1):
                time1 = 1.0 * step / n
                time2 = 1.0 * (step+1) / n
                window_draw_line(curve(time1), curve(time2))

# These verisons of draw_curve also work, but use "while loops". We have
# not introduced while loops in class. 
# 
# def draw_curve_points(curve,n):
#     t=0
#     step = 1.0/n
#     while t<1.0:
#         window_draw_point(curve(t))
#         t = t+step
# 
# def draw_curve_connected(curve,n):
#     t=0
#     step = 1.0/n
#     while t<1.0-step:
#         window_draw_line(curve(t),curve(t+step))
#         t = t+step

# Some simple curves
def mid_line(t):
    return make_point(t, .5)

def horiz_line(t):
    return make_point(t,0)

def vertical_line(t):
    return make_point(0,t)

def unit_circle(t):
    return make_point(math.cos(math.pi*2*t),math.sin(math.pi*2*t))

# Functions for transforming curves into new curves.
def translate(curve,x,y):
    def new_curve(t):
        ct = curve(t)
        return make_colored_point(x+x_of_point(ct),y+y_of_point(ct),color_of_point(ct))
    return new_curve

def rotate_ccw(curve):
    def new_curve(t):
        ct = curve(t)
        return make_colored_point(y_of_point(ct),x_of_point(ct),color_of_point(ct))
    return new_curve

def flip_vertically(curve):
    def new_curve(t):
        ct = curve(t)
        return make_colored_point(x_of_point(ct)*-1,y_of_point(ct),color_of_point(ct))
    return new_curve

def shrink(curve,scale):
    def new_curve(t):
        ct = curve(t)
        return make_colored_point(scale*x_of_point(ct),scale*y_of_point(ct),color_of_point(ct))
    return new_curve

def first_half(curve):
    return lambda(t) : curve(t/2)

def degrees_to_radians(degrees):
    return (math.pi*degrees)/180

#
# rotate-around-origin counterclockwise by theta degrees
# (You do not need to understand the geometry math used here.) 
#
def rotate_around_origin(curve,theta):
    cth = math.cos(degrees_to_radians(theta))
    sth = math.sin(degrees_to_radians(theta))
    def new_curve(t):
        ct = curve(t)
        x = x_of_point(ct)
        y = y_of_point(ct)
        return make_colored_point(x*cth-y*sth,x*sth+y*cth,color_of_point(ct))
    return new_curve

#
# Scale a curve
#
def scale_x_y(curve,x_scale,y_scale):
    def new_curve(t):
        ct = curve(t)
        return make_colored_point(x_scale*x_of_point(ct),y_scale*y_of_point(ct),color_of_point(ct))
    return new_curve

def scale(curve,s):
    return scale_x_y(curve,s,s)

# squeeze-rectangular-portion translates and scales a curve
# so the portion of the curve in the rectangle
# with corners xlo xhi ylo yhi will appear in a display window
# which has x, y coordinates from 0 to 1.
def squeeze_rectangular_portion(curve,xlo,xhi,ylo,yhi):
    return scale_x_y(translate(curve,-1*xlo,-1*ylo,1.0/(xhi-xlo),1.0/(yhi-ylo)))

#
# connect-rigidly makes a curve consisting of curve1 followed by curve2.
#

def connect_rigidly(curve1, curve2):
    def new_curve(t):
        if(t < 0.5):
            return curve1(2*t)
        else:
            return curve2(2*t - 1)
    return new_curve

#
# (connect-curves-evenly curvelist)
# evaluates to a single curve made by connecting all the curves in
# curvelist in a way that will distribute all the t values evenly between
# all the curves.
#

def connect_curves_evenly(curvelist):
    def new_curve(t):
        if t>=1.0:
            which_curve = len(curvelist)-1
        else:
            which_curve = int(math.floor(t*len(curvelist)))
        return curvelist[which_curve](len(curvelist)*(t-which_curve*1.0/len(curvelist)))
    return new_curve

def construct_simple_curvelist(curve,curvelist):
    endpoint = curve(1.0)
    return [curve] + map(lambda thiscurve : translate(thiscurve,x_of_point(endpoint),y_of_point(endpoint)),curvelist)

#
# These procedures find the extents of a curve, so we can scale it to the
# window:
#
def find_extreme_point(curve,point_selector,comparison,n):
    t = 0.0
    step = 1.0/n
    best_so_far = False
    while t<1.0:
        if ((not best_so_far) or comparison(point_selector(curve(t)),best_so_far)):
            best_so_far = point_selector(curve(t))
        t = t + step
    if comparison(point_selector(curve(1.0)),best_so_far):
        return point_selector(curve(1.0))
    else:
        return best_so_far

def find_leftmost_point(curve,n):
    return find_extreme_point(curve,x_of_point,lambda a,b:a<b,n)
def find_rightmost_point(curve,n):
    return find_extreme_point(curve,x_of_point,lambda a,b:a>b,n)
def find_lowest_point(curve,n):
    return find_extreme_point(curve,y_of_point,lambda a,b:a<b,n)
def find_highest_point(curve,n):
    return find_extreme_point(curve,y_of_point,lambda a,b:a>b,n)

def position_curve(curve,startx,starty):
    tcurve = translate(curve,startx,starty)
    num_points = 1000
    xlo = find_leftmost_point(tcurve,num_points)
    xhi = find_rightmost_point(tcurve,num_points)
    ylo = find_lowest_point(tcurve,num_points)
    yhi = find_highest_point(tcurve,num_points)
    if xlo < 0.01:
        xlowscale = (startx-0.01)/(startx-xlo)
    else:
        xlowscale = 1.0
    if xhi > 0.99:
        xhighscale = (0.99-startx)/(xhi-startx)
    else:
        xhighscale = 1.0
    if ylo < 0.01:
        ylowscale = (starty-0.01)/(starty-ylo)
    else:
        ylowscale = 1.0
    if yhi > 0.99:
        yhighscale = (0.99-starty)/(yhi-starty)
    else:
        yhighscale = 1.0
    minscale = min(xlowscale,xhighscale,ylowscale,yhighscale)
    return translate(scale_x_y(curve,minscale,minscale),startx,starty)

def save_image(filename):
        from canvasvg import saveall
        turtle.hideturtle() 
        ts = turtle.getscreen()
        canvas = ts.getcanvas()
        saveall(filename, canvas) 
        # ts.getcanvas().postscript(file=filename) 

